const { JSONPath } = require('jsonpath-plus');
const Output = require('./Output.js');

class Parser {

  constructor(json) {
    this.json = "";
    if (typeof json === "string") {
      this.json = JSON.parse(json);
    } else {
      this.json = json;
    }
    this.output = new Output();
  }

  parse() {
    // meta info
    var version = JSONPath({ path: "$.info.version", json: this.json });
    var title = JSONPath({ path: "$.info.title", json: this.json });
    this.output.addMeta(version, title);

    // get paths
    var paths = JSONPath({ path: "$.paths.*~", json: this.json });

    paths.forEach(path => {
      this.parsePath(path); //eg: /auth/login
    });

    return this.output.getAbapCode();
  }


  parsePath(path) {
    // get operations
    var opertions = JSONPath({ path: "$.paths." + path + ".*~", json: this.json });

    opertions.forEach(operation => {
      this.parseOperation(path, operation); //eg: POST
      this.output.finishPath();
    })
  }


  parseOperation(path, operation) {
    var summary = JSONPath({ path: "$.paths." + path + "." + operation + ".summary", json: this.json });
    var description = JSONPath({ path: "$.paths." + path + "." + operation + ".description", json: this.json });

    // operationID leer -> ERROR
    var operationId = JSONPath({ path: "$.paths." + path + "." + operation + ".operationId", json: this.json });
    if (!operationId || operationId.length < 1) {
      throw 'ERROR: operationId of path: "' + path + '" missing';
    }

    // x-weco-request fehlt -> ERROR
    var xWecoRequest = JSONPath({ path: "$.paths." + path + "." + operation + ".x-weco-request", json: this.json });
    if (!xWecoRequest) {
      throw 'ERROR: x-weco-request of path: "' + path + '" missing';
    }

    var security = JSONPath({ path: "$.paths." + path + "." + operation + ".security.~", json: this.json });
    security = security[2];

    this.output.addPath(path, operation, operationId, xWecoRequest, security);

    // get parameters
    var parameters = JSONPath({ path: "$.paths." + path + "." + operation + ".parameters.*", json: this.json });

    parameters.forEach(parameter => {
      if (parameter.$ref) {
        parameter = this.resolveRef(parameter.$ref);
      }

      this.parseParameter(parameter);
    });

    // get request body
    var requestBody = JSONPath({ path: "$.paths." + path + "." + operation + ".requestBody", json: this.json });
    if (requestBody.length == 1) {

      // resolve $ref
      if (requestBody[0].$ref) {
        this.parseRequestBodyByRef(requestBody[0].$ref);
      } else {
        this.parseRequestBody(path, operation, requestBody[0]);
      }
    }
  }

  parseRequestBodyByRef(ref) {
    var body = this.resolveRef(ref);

    var required = body.required;
    var type = Object.getOwnPropertyNames(body.content);
    type = type[0];

    this.output.addRequestBody(required, type);

    // Request Body Properties

    // schema with ref
    if(body.content[type].schema.$ref) {
      var schema = this.resolveRef(body.content[type].schema.$ref);
      this.addRequestBodyFromSchema(schema);
    }
    // schema
    else {
      var schema = {
        required: body.content[type].schema.required || [],
        properties: body.content[type].schema.properties
      }

      this.addRequestBodyFromSchema(schema);
    }

    this.output.finishRequestBody();
  }

  /**
   * 
   * @param {
   *   required: required,
   *   properties: properties
   * } schema 
   */
  addRequestBodyFromSchema(schema) {
    var requiredProperties = schema.required || [];
    var properties = schema.properties;
    var propertyNames = Object.getOwnPropertyNames(properties);

    propertyNames.forEach(propertyName => {
      var propertyType = properties[propertyName].type; 

      var propertyEnum = properties[propertyName].enum;
      propertyEnum = propertyEnum || [];

      var propertyRequired = requiredProperties.includes(propertyName);
      this.output.addRequestBodyProperty(propertyName, propertyType, propertyRequired, propertyEnum);
    });  
  }

  parseRequestBody(path, operation, requestBody) {
    // Request Body
    var required = requestBody.required;
    var type = JSONPath({ path: "$.paths." + path + "." + operation + ".requestBody.content.*~", json: this.json });
    this.output.addRequestBody(required, type[0]);

    // Request Body Properties

    // schema with ref
    var ref = JSONPath({ path: "$.paths." + path + "." + operation + ".requestBody.content." + type + ".schema.$ref", json: this.json });
    if(ref && ref.length > 0) {
      ref = ref[0];
      var schema = this.resolveRef(ref);

      this.addRequestBodyFromSchema(schema);
    }
    // schema
    else {
      var requiredProperties = JSONPath({ path: "$.paths." + path + "." + operation + ".requestBody.content." + type + ".schema.required.*", json: this.json });
      var propertyNames = JSONPath({ path: "$.paths." + path + "." + operation + ".requestBody.content." + type + ".schema.properties.*~", json: this.json });
  
      propertyNames.forEach(propertyName => {
        var propertyType = JSONPath({
          path: "$.paths." + path + "." + operation + ".requestBody.content."
            + type + ".schema.properties." + propertyName + ".type", json: this.json
        });
        propertyType = propertyType[0];
  
        var propertyEnum = JSONPath({
          path: "$.paths." + path + "." + operation + ".requestBody.content."
            + type + ".schema.properties." + propertyName + ".enum", json: this.json
        });
        propertyEnum = propertyEnum[0] || [];
  
        var propertyRequired = requiredProperties.includes(propertyName);
        this.output.addRequestBodyProperty(propertyName, propertyType, propertyRequired, propertyEnum);
      });
    }

    this.output.finishRequestBody();
  }


  parseParameter(parameter) {
    var name = parameter.name;
    var inSegment = parameter.in;
    var required = parameter.required ? 'X' : '';
    var enumValues = [];

    if (parameter.schema) {
      if (parameter.schema.type) {
        var type = parameter.schema.type;
      }
      if (parameter.schema.enum) {
        enumValues = parameter.schema.enum;
      }
    }

    this.output.addParameter(name, inSegment, required, type, enumValues);
  }

  resolveRef(ref) {
    var path = ref;
    path = this.replaceAll(path, '#/', '');
    path = this.replaceAll(path, '/', '.');

    var result = JSONPath({ path: "$." + path, json: this.json });
    result = result[0];

    if(result.$ref) {
      return this.resolveRef(result.$ref);
    }
    return result;
  }

  replaceAll(string, search, replacement) {
    if(!string) {
      return;
    }
    return string.replace(new RegExp(search, 'g'), replacement);
  }

}

module.exports = Parser;